#ifdef HAVE_CONFIG_H
#include "../config.h"
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <utmp.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <protocols/rwhod.h>
#include "rwhod.h"

#ifndef HAVE_GETLOADAVG
extern int getloadavg(double *, int);
#endif

#ifdef __linux__
/*
 * Do the one-time initialisations for this struct
 */
int initwhod(struct whod *rwho)
{
	FILE	*fp;
	float	f_uptime;
	char	hname[256];
	char	*p;
	time_t	now;

	memset(rwho,0,sizeof(struct whod));

	fp = fopen("/proc/uptime","r");
	if (fp == NULL)
		return 0;
	if (fscanf(fp,"%f",&f_uptime) != 1) {
		fclose(fp);
		return 0;
	}
	fclose(fp);
	time(&now);
	rwho->wd_boottime = now - f_uptime;
	rwho->wd_boottime = htonl(rwho->wd_boottime);

	if (gethostname(hname,sizeof(hname)) != 0)
		return 0;
	p = strchr(hname,'.');
	if (p) *p = '\0';

	/* now place the fixed members in the rwhod struct */
	rwho->wd_vers = WHODVERSION;
	rwho->wd_type = WHODTYPE_STATUS;
	strcpy(rwho->wd_hostname,hname);

	return 1;
}

/*
 * Do the runtime updates for this struct
 */
int updatewhod(struct whod *rwho)
{
	double		f_la[3];
	struct	utmp	*pu;
	struct	whoent	*wp = rwho->wd_we;
	time_t		now;

	/* gather some data */
	getloadavg(f_la,3);
	time(&now);

	while ((pu=getutent()) != NULL) {
		char		tmp[80];
		struct	stat	st;

		if (pu->ut_type != USER_PROCESS)
			continue;

		strcat(strcpy(tmp,"/dev/"),pu->ut_line);
		stat(tmp,&st);
		wp->we_idle = htonl(now - st.st_atime);
		memcpy(wp->we_utmp.out_line,pu->ut_line,8);
		memcpy(wp->we_utmp.out_name,pu->ut_user,8);
		wp->we_utmp.out_time = htonl(pu->ut_time);

		wp++;
	}
	endutent();

	/* convert to network formats */
	rwho->wd_sendtime = htonl(now);
	rwho->wd_loadav[0] = f_la[0] * 100;
	rwho->wd_loadav[1] = f_la[1] * 100;
	rwho->wd_loadav[2] = f_la[2] * 100;
	rwho->wd_loadav[0] = htonl(rwho->wd_loadav[0]);
	rwho->wd_loadav[1] = htonl(rwho->wd_loadav[1]);
	rwho->wd_loadav[2] = htonl(rwho->wd_loadav[2]);

	return (char *)wp - (char *)rwho;
}
#endif

#ifdef __sun__

#include <fcntl.h>
#include <nlist.h>
#include <sys/time.h>

static	struct	nlist	nl[] = { { "_boottime" }, { "_avenrun" }, { NULL } };
static	int	memfd;

/*
 * Do the one-time initialisations for this struct
 */
int initwhod(struct whod *rwho)
{
	struct timeval	boottime;
	char	hname[256];
	char	*p;

	memset(rwho,0,sizeof(struct whod));

	if (nl[0].n_value == 0) {
		nlist("/vmunix",&nl);
		if (nl[0].n_name[0] == '\0' || nl[1].n_name[0] == '\0') {
			fprintf(stderr,"%s: unable to locate symbols\n",progname);
			return 0;
		}
	}

	/* fetch boottime */
	memfd = open("/dev/kmem",O_RDONLY);
	if (memfd<0) {
		perror(progname);
		return 0;
	}
	lseek(memfd,nl[0].n_value,SEEK_SET);
	read(memfd,&boottime,sizeof(boottime));
	rwho->wd_boottime = htonl(boottime.tv_sec);

	gethostname(hname,sizeof(hname));
	p = strchr(hname,'.');
	if (p) *p = '\0';

	/* now place the fixed members in the rwhod struct */
	rwho->wd_vers = WHODVERSION;
	rwho->wd_type = WHODTYPE_STATUS;
	strcpy(rwho->wd_hostname,hname);

	rwho->wd_boottime = htonl(rwho->wd_boottime);
}

/*
 * Do the runtime updates for this struct
 */
int updatewhod(struct whod *rwho)
{
	int		fd;
	double		f_la[3];
	struct	utmp	ut;
	struct	whoent	*wp = rwho->wd_we;
	time_t		now;

	/* gather some data */
	getloadavg(f_la,3);
	time(&now);

	fd = open("/etc/utmp",O_RDONLY);
	if (fd >= 0) {
		while (read(fd,&ut,sizeof(ut)) == sizeof(ut)) {
			char		tmp[32];
			struct	stat	st;

			/* empty entry? */
			if (ut.ut_name[0]=='\0' && ut.ut_host[0]=='\0'
			    || nonuser(ut))
				continue;

			strcat(strcpy(tmp,"/dev/"),ut.ut_line);
			stat(tmp,&st);
			wp->we_idle = htonl(now - st.st_atime);
			memcpy(wp->we_utmp.out_line,ut.ut_line,8);
			memcpy(wp->we_utmp.out_name,ut.ut_name,8);
			wp->we_utmp.out_time = htonl(ut.ut_time);

			wp++;
		}
		close(fd);
	}

	/* translate to network data */
	rwho->wd_sendtime = htonl(now);
	rwho->wd_loadav[0] = htonl(f_la[0] * 100);
	rwho->wd_loadav[1] = htonl(f_la[1] * 100);
	rwho->wd_loadav[2] = htonl(f_la[2] * 100);

	return (char *)wp - (char *)rwho;
}
#endif
